import pandas as pd
import pysquid.templating
from pysquid.core import _ID_TASK, _ID_VAR


def to_df(lu_):

    data_types = {
        'priority': [], 'uuid': []
    }

    for uuid, template in lu_.items():    
        data_types['priority'].append(template.get('priority'))
        data_types['uuid'].append(str(uuid))
    
    return pd.DataFrame(data_types).sort_values('priority')


def merge(templates: list = []):

    lu_ = {}
    
    for template in templates:

        if not isinstance(template, dict):
            continue

        d_ = pysquid.templating.parse_template(template)

        uuid = str(d_.get('__uuid__'))
        lu_[uuid] = d_

    df = to_df(lu_)

    grouped = df.groupby(['priority'])

    tags = set()
    untags = set()

    services = {}
    variables = {}
    
    for gid, group in grouped:    
        
        uuids = group.get('uuid').tolist()
        
        for uuid in uuids:
            
            template = lu_.get(str(uuid))
            
            tags_ = template.get('tags')
            untags_ = template.get('untags')
                
            tags = tags.union(tags_)
            untags = untags.union(untags_)

            services_ = template.get(_ID_TASK)
            variables_ = template.get(_ID_VAR)

            services = {**services, **services_}
            variables = {**variables, **variables_}
            
    parsed_template = {
        'tags': tags.difference(untags),
        _ID_TASK: services,
        _ID_VAR: variables,
        'global': {},
    }

    return parsed_template


def build_services(parsed_template: dict):

    services = parsed_template.get(_ID_TASK)

    SERVICE_MAGICS = {
        'plugin_id': (str, '', str, None),
        'tags': (list, [], set, None),
        _ID_VAR: (dict, {}, dict, None),
    }

    services_ = {}
    variables_ = parsed_template.get(_ID_VAR)

    for sid, service in services.items():
        parsed = pysquid.templating.parse_template(service, SERVICE_MAGICS)
        parsed['plugin_id'] = parsed['plugin_id'] or sid
        service_vars = variables_.get(sid) or {}
        parsed[_ID_VAR] = {**service_vars, **parsed.get(_ID_VAR)}
        services_[sid] = parsed

    parsed_template[_ID_TASK] = services


def merge_templates(templates: list = []):

    templates = sorted(templates, key=lambda k: k['priority']) 

    merged = {}

    for template in templates:
        merged = {**merged, **template}

    merged['__tags__'] = merged.get('tags').difference(merged.get('untags'))

    return merged



import yaml
import pathlib


def load_yaml(path: pathlib.Path) -> dict:
    """
    Load a YAML file
    """
    with open(path, 'r') as stream:
        return yaml.safe_load(stream)
        
    
def load_toml(path: pathlib.Path) -> dict:
    """
    Load a TOML file
    """
    return {}
#    with open(path, 'r') as stream:
#        return {}


def load_json(path: pathlib.Path) -> dict:
    """
    Load a JSON file
    """
    return {}
#    with open(path, 'r') as stream:
#        return {}



import logging


def stream_logger():
    # Simple logger when no provided
    logger = logging.getLogger('stream')
    logger.setLevel(logging.DEBUG)
    # create formatter and add it to the handlers
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(threadName)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(ch)
    return logger





class Worker():

    def __init__(self, id_: str, pool: int = 1, stage: int = 1, mode: str = 'thread'):
        self.id_ = id_
        self.pool = pool
        self.stage = stage

    def add_service(self, service, template):
        self.service = service
        self.template = template
        self.variables = self.service.get('variables')

    def pre(self):
        pass

    def post(self):
        pass
    
    def apply(self):
        print(self.service)


class Hook():

    def __init__(self):
        self.current_worker = None

    def set_worker(self, worker: Worker):
        self.current_worker = worker

    def init(self):
        pass

    def cleanup(self):
        pass
    
        
class Plugin():

    def __init__(self, plugin):
        self.plugin_id = plugin
        self.pools = set()
        self.stages = set()
        self.service = {}
        self.template = {}
        self.workers = {}
        self.enabled_workers = []

    def add_service(self, service, template):
        self.service = service
        self.template = template

    def init_hook(self, worker):
        pass
        
    def iterate_workers(self, enabled_workers: set = None):

        setup = {}
        
        keys = set(self.workers.keys())
        
        if enabled_workers:
            keys = keys.intersection(enabled_workers)

        for key in keys:
            worker = self.workers.get(key)
            w = worker()
            w.add_service(self.service, self.template)
            self.init_hook(w)
            self.enabled_workers.append(w)
            
        for worker in self.enabled_workers:

            pool = worker.pool
            stage = worker.stage

            if pool not in setup:
                setup[pool] = {}

            if stage not in setup[pool]:
                setup[pool][stage] = []

            setup[pool][stage].append(worker)
            
        return setup
            
    def apply(self):
        print(self.plugin_id)



import pysquid.plugin


class SampleWorker1(pysquid.plugin.Worker):
    def __init__(self):
        super().__init__('sample_worker_1')

        
class SampleWorker2(pysquid.plugin.Worker):
    def __init__(self):
        super().__init__('sample_worker_1', 1, 3)
        
    def apply(self):
        import time
        time.sleep(10)

        
class SampleWorker3(pysquid.plugin.Worker):
    def __init__(self):
        super().__init__('sample_worker_3', 1, 2)

    def apply(self):
        import time
        time.sleep(10)


class SampleWorker4(pysquid.plugin.Worker):
    def __init__(self):
        super().__init__('sample_worker_4')

    def apply(self):
        from pathlib import Path

        p = Path('test/e2e/out/simple_1')

        if p.exists():
            p.unlink()

        p.touch()

        
class SamplePlugin1(pysquid.plugin.Plugin):
    def __init__(self):
        super().__init__('sample_plugin_1')
        
        self.workers = {
            'sample_worker_1': SampleWorker1,
            'sample_worker_2': SampleWorker2,
            'sample_worker_3': SampleWorker3
        }

        
class SamplePlugin2(pysquid.plugin.Plugin):
    def __init__(self):
        super().__init__('sample_plugin_2')
        
        self.workers = {
            'sample_worker_4': SampleWorker4,
        }

        
EXPORTS = [
    SamplePlugin1,
    SamplePlugin2
]



from concurrent.futures import ThreadPoolExecutor, wait, Future
import threading
import sys
from pprint import pformat
from pysquid import stream_logger
from pysquid.plugin import Hook
from pysquid.core import _ID_TASK


def setup_thread_excepthook():
    """
    Workaround for sys.excepthook bug
    """

    init_orginal = threading.Thread.__init__

    def init(self, *args, **kwargs):

        init_orginal(self, *args, **kwargs)
        run_original = self.run

        def run_with_except_hook(*args2, **kwargs2):
            try:
                run_original(*args2, **kwargs2)

            except Exception:
                sys.excepthook(*sys.exc_info())

        self.run = run_with_except_hook

    threading.Thread.__init__ = init
    

class ThreadedEngine():

    def __init__(self, template, plugins, except_hook=None, log=None, max_workers=50, max_pool_workers=20):
        self.template = template
        self.plugins = plugins
        self.log = log if log else stream_logger()
        self.max_workers = max_workers
        self.max_pool_workers = max_pool_workers
        setup_thread_excepthook()
        sys.excepthook = except_hook if except_hook else self.inbuilt_except_hook
        self.exceptions = []
        self.pools = {}

    def inbuilt_except_hook(self, exctype, value, tb):
        """
        Catch exceptions
        """
        self.log.info(f'Exception: {exctype}, {value}')
        self.exceptions.append({'type': exctype})
        
    def build(self):

        services = self.template.get(_ID_TASK)
        plugins = self.plugins
        
        pools = {
            'thread': {}
        }

        playbook = {}

        for sid, service in services.items():

            mode = service.get('__mode__')
            plugin = service.get('__plugin__')
            workers = service.get('__workers__')

            if mode not in pools or plugin not in plugins:
                continue

            enabled_workers = set(workers.keys())

            plugin_ = plugins.get(plugin)()
            plugin_.add_service(service, self.template)

            setup = plugin_.iterate_workers(enabled_workers)

            for pid, stages in setup.items():
                if pid not in playbook:
                    playbook[pid] = {}
                    
                for sid, stage in stages.items():
                    if sid not in playbook[pid]:
                        playbook[pid][sid] = []

                    playbook[pid][sid] = playbook[pid][sid] + setup[pid][sid] 

        pools['thread'] = playbook
        self.pools = pools
        
    def exec_pool(self, hook: Hook = None):

        pools = self.pools.get('thread')        
        futures = set()

        hook = hook if isinstance(hook, Hook) else Hook()

        exception_count = 0
        
        for pid, pool in pools.items():
            self.log.info(f'Running pool {pid}')
            e = ThreadPoolExecutor(max_workers=self.max_pool_workers)

            try:            
                future = e.submit(self.exec_workers, pool, hook)

                if isinstance(future, Future):
                    futures.add(future)

            except Exception as _e:
                self.log.info(f'Error submitting pool: {_e!r}')

        self.log.info(f'Waiting on {futures}')
        wait(futures)

        for finished_worker in futures:
            self.log.debug(pformat(finished_worker.__dict__))
            exception_count += finished_worker._result
            
        self.log.info(f'Future done: {futures}, exceptions: {exception_count}')

        return exception_count
            
    def exec_workers(self, pool, hook):

        e = ThreadPoolExecutor(max_workers=self.max_workers)
        futures = set()

        pipeline = dict(sorted(pool.items(), key=lambda item: item[0]))

        exceptions = 0
        
        for sid, stage in pipeline.items():
            
            for worker in stage:
                # Set worker and run init

                try:
                    hook.set_worker(worker)
                    hook.init()                
                    worker.pre()
                    future = e.submit(worker.apply)

                    if isinstance(future, Future):
                        futures.add(future)

                except Exception as _e:
                    self.log.info(f'Error submitting worker: {_e!r}')

            wait(futures)

            for finished_worker in futures:
                self.log.debug(pformat(finished_worker.__dict__))

                if finished_worker._exception:
                    exceptions += 1
                    
            for worker in stage:

                try:                
                    worker.post()
                    hook.set_worker(worker)
                    hook.cleanup()

                except Exception as _e:
                    self.log.info(f'Error submitting worker: {_e!r}')
                
        return exceptions
    





import uuid
from functools import reduce
from copy import deepcopy
from pysquid.core import _ID_TASK, _ID_VAR


TEMPLATE_STRUCT = {
    'priority': (int, 0, int, None),
    'tags': (list, [], set, None),
    'untags': (list, [], set, None),
    _ID_VAR: (dict, {}, dict, None),
    'global': (dict, {}, dict, None),
    _ID_TASK: (dict, {}, dict, None),
}

TEMPLATE_INTERNALS = {
    '__uuid__': lambda k, v, id_: uuid.uuid4(),
}

SERVICE_STRUCT = {
    'tags': (list, [], set, None),
    _ID_VAR: (dict, {}, dict, None),
    '__matching__': (str, 'weak', str, None),  # weak | strong | exact
    '__enabled__': (bool, False, bool, None),  # weak | strong | exact
    '__mode__': (str, 'thread', str, None),
    '__workers__': (dict, {}, dict, None),
}

SERVICE_INTERNALS = {
    '__plugin__': lambda k, v, id_: v.get('__plugin__') if v.get('__plugin__') else id_,
    '__variable__': lambda k, v, id_: v.get('__variable__') if v.get('__variable__') else id_
}


def parse_template(d: dict, struct: dict, internals: dict, id_: str = None):
    
    for mkey, value in struct.items():

        type_, default_, converterf_, allowed_ = value

        value_from_template = d.get(mkey)

        if value_from_template and not isinstance(value_from_template, type_):
            raise ValueError

        current_ = value_from_template or default_

        if not isinstance(current_, converterf_):                    
            current_ = converterf_(current_)

        if allowed_ and current_ not in allowed_:
            raise ValueError

        d[mkey] = current_

    for mkey, func in internals.items():
        d[mkey] = func(mkey, d, id_)

    return d


def merge(a, b, path=None):
    "merges b into a"
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])            
        else:
            a[key] = b[key]
    return a


class Template():

    def __str__(self):
        import pprint
        return pprint.pformat(self.template)

    def __init__(self):
        self.template = {}
        self.subtemplates = []
        self.priorites = []
        self.tags = set()
        self.untags = set()
        self.task_struct = deepcopy(SERVICE_STRUCT)
        self.task_internal = deepcopy(SERVICE_INTERNALS)
        self.template_struct = deepcopy(TEMPLATE_STRUCT)
        self.template_internals = deepcopy(TEMPLATE_INTERNALS)

    def add_subtemplate(self, subtemplate: dict = None):

        subtemplate = subtemplate if subtemplate else {}

        d = parse_template(subtemplate, self.template_struct, self.template_internals)

        tags = d.get('tags')
        untags = d.get('untags')

        self.tags = self.tags.union(tags)
        self.untags = self.untags.union(untags)

        services = d.get(_ID_TASK)
        variables = d.get(_ID_VAR)
        
        for sid, service in services.items():
            d[_ID_TASK][sid] = parse_template(service, self.task_struct, self.task_internal, sid)
            variable_key = d[_ID_TASK][sid].get('__variable__')
            variables_ = merge(variables.get(variable_key, {}), d[_ID_TASK][sid].get(_ID_VAR))
            d[_ID_TASK][sid][_ID_VAR] = variables_

        self.subtemplates.append(d)

        self.priorites.append(d.get('priority'))
        self.priorites.sort()
        
        return d
    
    def build(self, match: bool = True):

        templates = sorted(self.subtemplates, key=lambda k: k['priority'], reverse=True) 

        reduce(merge, templates)

        merged = templates[0]
        
        services = merged.get(_ID_TASK)
        variables = merged.get(_ID_VAR)

        for sid, service in services.items():
            key = service.get('__variable__')
            # variables_ = merge(variables.get(key, {}), service.get(_ID_VAR))
            variables_ = merge(service.get(_ID_VAR), variables.get(key, {}))
            service[_ID_VAR] = variables_

        merged['__tags__'] = self.tags.difference(self.untags)

        self.template = merged

        if match:
            self.match()

        return self.template

    def match(self):

        services = self.template.get(_ID_TASK)
        tags = self.template.get('__tags__')

        matched_services = {}
        
        for sid, service in services.items():

            stags = service.get('tags')

            weak = tags.intersection(stags)
            strong = tags.difference(stags)

            matchmode = service.get('__matching__')

            if matchmode == 'strong' and not stags:
                continue
            
            if matchmode == 'weak' and (weak or not stags):
                matched_services[sid] = service
                continue

            if matchmode == 'strong' and strong:
                matched_services[sid] = service
                continue
                
        self.template[_ID_TASK] = matched_services

        return self.template
            




import yaml
from setuptools import setup, find_packages
from io import open
from os import path

readme = path.join(path.abspath(path.dirname(__file__)), 'README.md')
with open(readme, 'r') as f:
    readme_contents = f.read()

with open('build.yaml', 'r') as stream:
    config = yaml.safe_load(stream)

name = config.get('project')
version = config.get('version')
description = config.get('description')
author = config.get('author')
author_email = config.get('author_email')
url = config.get('url')

install_requires = [
    'pyyaml',
]    
  
setup(
    name=name,
    version=version,
    description=description,
    long_description=readme_contents,
    long_description_content_type='text/markdown',
    url=url,
    author=author,
    author_email=author_email,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='python development library',
    packages=find_packages(),
    install_requires=install_requires,
    python_requires='>=3.8',
    project_urls={
        'Bug Reports': url,
        'Tracker': url,
        'Source': url,
    },
)



from pathlib import Path
import importlib
import glob
import pysquid.plugin


class PluginCollector():

    def __init__(self, paths: list = [Path('plugins')]):

        self.paths = set()
        self.pyfiles = set()
        self.pymodules = set()
        self.plugins = {}
        self.get_paths(paths)
        self.get_py_files()

    def get_paths(self, paths):
        
        for path in paths:

            if not isinstance(path, Path):
                continue

            if not path.exists():
                continue

            if not path.is_dir():
                continue

            self.paths.add(path)

    def get_py_files(self):

        all_files = set()

        for path in self.paths:
            files = set(glob.glob(str(path) + '/**/*.py', recursive=True))
            all_files = all_files.union(files)

        self.pyfiles = all_files

        files = [f.replace('.py', '') for f in all_files]
        self.add_modules([f.replace('/__init__', '') for f in files])

    def add_modules(self, modules):
        self.pymodules = self.pymodules.union(set([f.replace('/', '.') for f in modules]))

    def add_plugin(self, plugin):

        try:
            ti = plugin()

            if isinstance(ti, pysquid.plugin.Plugin):
                self.plugins[ti.plugin_id] = plugin

        except Exception as e:
            msg = f'Error adding plugin: {e!r}'
            print(msg)

    def add_plugins(self, plugins: list = []):

        for plugin in plugins:
            self.add_plugin(plugin)

    def collect(self):

        for module in self.pymodules:

            try:
                imported_module = importlib.import_module(module)

                exports = []

                try:
                    exports = imported_module.EXPORTS
                except Exception:
                    pass

                for export in exports:

                    self.add_plugin(export)

            except Exception as e:
                msg = f'Error importing module: {module}: {e!r}'
                print(msg)

        
        
            
